#' Removal of probes from a Chip Definition File (CDF).
#'
#' \code{cdf.sniper} rebuilds a CDF using only whitelisted probes.
#'
#' Write some text here about the format of the CDF, where these can be obtained
#' including links. The format of the whitelist should also be detailed (as
#' returned by \code{\link[CDFSnipeR]{remove.probes}}).
#'
#' @param cdf.file The CDF which is to be manipulated. This can also be a connection (or URL, anything that is readable by utils::read.table).
#' @param clean.probes A whitelist of probe coordinates to be retained.
#' @param min.probeset.size The minimum number of probes in a probset.
#' If fewer than this number remain after probe removal, the entire probset will be discarded.
#' @param keep.affx A logical flag indicating whether the AFFX control probes should be retained.
#' @param newline It may be necessary to change this on non-UNIX-like operating systems.
#' @param cluster An object of class 'cluster' as provided by the package 'parallel'. If this is provided, the
#' computationally expensive part of the code can be distributed over multiple CPUs.
#' @return A string containing the CDF text. It is intended that this is written
#' to a file using \code{\link[base]{write}}.
#'
#' @examples
#'
#' \dontrun{
#' newProbes <- remove.probes(oldProbes, snpPositions)
#' newCdf <- cdf.sniper(oldCdf, newProbes$FilteredProbes)
#' write(newCdf, file="newCdf.cdf")
#' }
#'
#' @export
cdf.sniper = function(cdf.file, clean.probes, min.probeset.size=3, keep.affx=TRUE, newline="\n", cluster=NULL){
    timer = system.time({
    #@ keep.affx sets whether to remove AFFX control probesets
    ## Read the CDF as a list of positions
    cdf.text = as.character(utils::read.table(cdf.file, sep="~")[, 1]) # Separator character should not exist in text
    section.breaks = grep("\\[", cdf.text)
    section.segments = cbind(c(1, section.breaks[2:length(section.breaks)]), c(section.breaks[2:length(section.breaks)]-1, length(cdf.text)))
    ## Break up into units
    #@ Matrix to store unit summary info (will be updated after probeset formatting completed)
    #@ All except Index are set when the probesets are being reformatted
    #@ Index is the row number of section.segments where the updated string needs to be written (set when we hit that row)
    section.pointers = data.frame("SectionName"=character(length=nrow(section.segments)), "UnitIndex"=numeric(length=nrow(section.segments)), "UnitDefStart"=numeric(length=nrow(section.segments)), "UnitDefEnd"=numeric(length=nrow(section.segments)), "BlockStart"=numeric(length=nrow(section.segments)), "BlockEnd"=numeric(length=nrow(section.segments)), "UnitName"=character(length=nrow(section.segments)), "NumAtoms"=numeric(length=nrow(section.segments)), "NumCells"=numeric(length=nrow(section.segments)), "ProbesetWritten"=logical(length=nrow(section.segments)), stringsAsFactors=F)
    for(segment.index in 1:nrow(section.segments)){
      section.code = cdf.text[section.segments[segment.index, 1]]
      if(length(grep("\\[Unit[0-9]*\\]", section.code))==1){
        section.pointers[segment.index, c("SectionName")] = section.code
        unit.id = as.numeric(sub("\\]", "", sub("\\[Unit", "", section.code)))
        section.pointers[segment.index, c("UnitIndex")] = unit.id
        section.pointers[segment.index, c("UnitDefStart", "UnitDefEnd")] = section.segments[segment.index, ]
      }else if(length(grep("\\[Unit[0-9]*\\_Block[0-9]*\\]", section.code))==1){
        unit.id = as.numeric(sub("\\_Block[0-9]*\\]", "", sub("\\[Unit", "", section.code)))
        pointer.index = which(section.pointers[, c("UnitIndex")]==unit.id) # Assume that Unit has already been defined
        section.pointers[pointer.index, c("BlockStart", "BlockEnd")] = section.segments[segment.index, ]
      }
      else{
        section.pointers[segment.index, c("SectionName")] = section.code
        section.pointers[segment.index, c("UnitDefStart", "UnitDefEnd")] = section.segments[segment.index, ]
      }
    }
    section.pointers = section.pointers[which(section.pointers[, "UnitDefStart"] > 0), ]
    row.names(section.pointers) = section.pointers$SectionName

    ## Break up data into sections
    #@ Unit sections will be fed to a loop for probe removal and header updating
    #@ The remainder remain untouched except for 'Chip' which has summary info updated
    other.sections = lapply(which(section.pointers[, "UnitIndex"] == 0), function(section.index){
      section = cdf.text[section.pointers[section.index, "UnitDefStart"]:section.pointers[section.index, "UnitDefEnd"]]
      return(section)
    })
    unit.section.list = lapply(which(section.pointers[, "UnitIndex"] > 0), function(section.index){
      unit.definition = cdf.text[section.pointers[section.index, "UnitDefStart"]:section.pointers[section.index, "UnitDefEnd"]]
      unit.block = cdf.text[section.pointers[section.index, "BlockStart"]:section.pointers[section.index, "BlockEnd"]]
      return(list("index"=section.index, "pointers"=section.pointers[section.index, ], "definition"=unit.definition, "block"=unit.block))
    })
    rm(cdf.text) # Free up memory

    ## An lapply loop makes this step easily parallelisable
    section.filter = function(unit){
      #@ First parse the block name
      block.name = strsplit(unit$block[grep("Name", unit$block)], "\\=")[[1]][2]
      if(length(grep("AFFX", block.name))==1) AFFX = TRUE else AFFX = FALSE # Is this an AFFX control probeset?
      #@ Then read the block into a data frame
      data.header.index = grep("CellHeader", unit$block) # This is the line that the Cell data starts
      data.table = utils::read.table(text=paste(unit$block[data.header.index:length(unit$block)], collapse="\n"), allowEscapes=T, header=T, sep="\t")
      #@ If an AFFX probeset, do nothing...
      if(AFFX){
        if(!keep.affx){ # ...unless AFFX probes are to be discarded.
          data.table = NULL
        }
      }else{ #@ Otherwise run the probe filtering
        present = apply(data.table, 1, function(row){
          x = as.numeric(sub("Cell[0-9]*\\=", "", row[1]))
          y = as.numeric(row[2])
          length(which(clean.probes$Probe.X==x & clean.probes$Probe.Y==y)) > 0
        })
        present.MM = data.table[, "EXPOS"] %in% data.table[present, "EXPOS"] # Mismatch probes are not in the list, but have same EXPOS value as the PM probe
        data.table = data.table[present.MM, ]
      }
      #@ Renumber the ATOMS and EXPOS columns with new sequential numbering
      atoms = (1:length(unique(data.table[, "ATOM"]))-1)
      names(atoms) = unique(data.table[, "ATOM"])
      data.table[, "ATOM"] = atoms[as.character(data.table[, "ATOM"])]
      data.table[, "EXPOS"] = atoms[as.character(data.table[, "ATOM"])]

      #@ Now rewrite the block definition info accordingly
      block.header = unname(sapply(unit$block[1:(data.header.index-1)], function(line){
        tag = strsplit(line, "\\=")[[1]][1]
        if(tag=="NumAtoms") line = paste0("NumAtoms=", max(atoms))
        else if(tag=="NumCells") line = paste0("NumCells=", max(0, nrow(data.table)))
        else if(tag=="StartPosition") line = paste0("StartPosition=", min(atoms))
        else if(tag=="StopPosition") line = paste0("StopPosition=", max(atoms))
        return(line)
      }))

      #@ Now the filtered probeset information can be used to update the unit definition
      unit.header = unname(sapply(unit$definition, function(line){
        tag = strsplit(line, "\\=")[[1]][1]
        if(tag=="NumAtoms") line = paste0("NumAtoms=", max(atoms))
        else if(tag=="NumCells") line = paste0("NumCells=", max(0, nrow(data.table)))
        return(line)
      }))

      #@ Add the probeset info to the register
      unit$pointers$NumAtoms = max(atoms)
      unit$pointers$NumCells = max(0, nrow(data.table))

      #@ Return the filtered unit as a character vector ready for rewriting
      if(length(atoms) < min.probeset.size){ # Are there sufficient probes to make a probeset?
        unit$pointers$ProbesetWritten = FALSE
        return(list("text"="", "pointers"=unit$pointers)) # Don't write anything
      }else{
        unit$pointers$ProbesetWritten = TRUE
        data.table$REGION = ""
        block = apply(data.table, 1, paste, collapse="\t")
        block = c(sub("CellHeader\\.X", "CellHeader\\=X", paste(colnames(data.table), collapse="\t")), block)
        return(list("text"=c(unit.header, newline, block.header, block, newline), "pointers"=unit$pointers))
      }
    }

    filtered.unit.section.list = NULL
    if(!sum(grep("cluster", class(cluster)))){
      filtered.unit.section.list= lapply(unit.section.list, section.filter)
    }else{
      parallel::clusterExport(cluster, list("clean.probes", "min.probeset.size", "keep.affx", "newline"), environment())
      filtered.unit.section.list = parallel::parLapply(cluster, unit.section.list, section.filter)
    }
    parallel::stopCluster(cluster)

    ## Summary results
    #@ Collate pointers
    filtered.pointers = do.call("rbind", lapply(filtered.unit.section.list, function(u) u$pointers ))
    section.pointers[filtered.pointers$SectionName, ] = filtered.pointers
    #@ Update summary data based on information collected during reformatting
    chip.section.index = which(section.pointers$SectionName=="[Chip]")
    other.sections[[chip.section.index]] = unname(sapply(other.sections[[chip.section.index]], function(line){
      tag = strsplit(line, "\\=")[[1]][1]
      if(tag=="NumberOfUnits") line = paste0("NumberOfUnits=", length(which(section.pointers[section.pointers$ProbesetWritten, "UnitIndex"]>0)))
      if(tag=="MaxUnit") line = paste0("MaxUnit=", max(section.pointers[section.pointers$ProbesetWritten, "UnitIndex"]))
      return(line)
    }))

    #@ Combine all text into a single vector for writing
    cdf.text = do.call("c", c(lapply(other.sections, function(x) c(x, newline)), lapply(filtered.unit.section.list, function(x) c(x$text, newline))))
  })
  print(timer)
  return(cdf.text)
}

##########

