#' Removal of probes from a probe list.
#'
#' \code{remove.probes} removes probes containing undesired features from a CDF.
#'
#' @param probes A data frame containing probe sequence information.
#' @param features A data frame containing genomic coordinates of undesired
#' features.
#' This should be in a BED-like format with the first three columns containing
#' the chromosome (with or without the 'chr'), the start position (in
#' nucleotides) and the end position (the same as the start in the case of SNPs).
#' @param probe.length A number specifying the length of the probes on the chip. This is 25 by default.
#' @return A list containing the following components:
#' FilteredProbes: A character vector of the whitelisted probes (i.e. those not
#' containing
#' the features defined in the BED file).
#' ProbeCounts: a matrix with one row for each probeset and two columns
#' containing the numbers of probes in each probeset both before (OldProbes) and
#' after (CleanProbes) probe removal.
#' RemovedProbes: A character vector of the removed probes (those containing
#' the features defined in the BED file).
#'
#' @examples
#'
#' \dontrun{
#' whitelist <- remove.probes(testProbes, testfeatures)
#' newCdf <- cdf.sniper(cdf, whitelist)
#' write(newCdf, file="newCdf_deSNP.cdf")
#' }
#'
#' @export
remove.probes = function(probes, features, probe.length = 25){
  ## Create a new list of probes not containing variant features
	probes$Chr = sub("Chr|chr", "", probes$Chr)
	features[, 1] = sub("Chr|chr", "", features[, 1])
  probe.ranges = GenomicRanges::GRanges(probes$Chr, IRanges::IRanges(probes$Chr.From, probes$Chr.From + probe.length - 1), probes$Chr.Strand)
  feature.ranges = GenomicRanges::GRanges(features[, 1], IRanges::IRanges(features[, 2], features[, 3]), "*")
  hits = IRanges::findOverlaps(feature.ranges, probe.ranges)
  clean.probes = probes[-S4Vectors::to(hits), ]
  ## Diagnostics on filtering
  probes.per.probeset = as.data.frame(table(probes$Probe.Set.Name), row.names=1)
  clean.probes.per.probeset = as.data.frame(table(clean.probes$Probe.Set.Name), row.names=1)
  probeset.change = merge(probes.per.probeset, clean.probes.per.probeset, by="row.names", all.x=T)
  row.names(probeset.change) = probeset.change$Row.names
  probeset.change = probeset.change[, -1]
  colnames(probeset.change) = c("OldProbes", "CleanProbes")
  return(list("FilteredProbes"=clean.probes, "ProbeCounts"=probeset.change, "RemovedProbes"=probes[S4Vectors::to(hits), ]))
}
